/*  Copyright (C) 1996-1997  Id Software, Inc.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    See file, 'COPYING', for details.
*/

/*
 * tomlite.c
 * Modifications by Kevin Shanahan, 1999-2000
 */

#include "tomlite.h"

float  scaledist      = 1.0;
float  scalecos       = 0.5;
float  rangescale     = 0.5;
int    worldminlight  = 0;
vec3_t minlight_color = { 255, 255, 255 }; /* defaults to white light   */
int    sunlight       = 0;
vec3_t sunlight_color = { 255, 255, 255 }; /* defaults to white light   */
vec3_t sunmangle      = { 0, 0, 16384 };   /* defaults to straight down */

byte     *filebase;
byte     *file_p;
byte     *file_end;

dmodel_t *bspmodel;
int      bspfileface;  /* next surface to dispatch */

vec3_t   bsp_origin;

qboolean extrasamples;
qboolean compress_ents;
qboolean facecounter;
qboolean colored;
qboolean tint;
qboolean nominlimit;

qboolean nolightface[MAX_MAP_FACES];
vec3_t   faceoffset[MAX_MAP_FACES];

byte *GetFileSpace (int size)
{
    byte *buf;

    LOCK;
    file_p = (byte *)(((long)file_p + 3)&~3);
    buf = file_p;
    file_p += size;
    UNLOCK;
    if (file_p > file_end)
        Error ("GetFileSpace: overrun");
    return buf;
}


void LightThread (void *junk)
{
    int i;

    while (1) {
        LOCK;
        i = bspfileface++;
        UNLOCK;
        if (!facecounter)
            printf("Lighting face %i of %i\r", i, numfaces);
        if (i >= numfaces) {
            logprint("\nLighting Completed.\n\n");
            return;
        }

		// Tomaz - Lit Begin
		if (!colored)
	        LightFace (i, nolightface[i], faceoffset[i]);
		else 
			LightFaceLIT (i, nolightface[i], faceoffset[i]);
		// Tomaz - Lit End
    }
}

void FindFaceOffsets( void )
{
    extern   int      nummodels;
    extern   dmodel_t dmodels[MAX_MAP_MODELS];
    int      i,j;
    entity_t *ent;
    char     name[ 20 ];
    char     *classname;
    vec3_t   org;

    memset( nolightface, 0, sizeof( nolightface ) );

    for( j = dmodels[ 0 ].firstface; j < dmodels[ 0 ].numfaces; j++ )
        nolightface[ j ] = 0;

    for( i = 1; i < nummodels; i++ ) {
        sprintf( name, "*%d", i );
        ent = FindEntityWithKeyPair( "model", name );
        if ( !ent )
            Error("FindFaceOffsets: Couldn't find entity for model %s.\n",
                  name );

        classname = ValueForKey ( ent, "classname" );
        if ( !strncmp( classname, "rotate_", 7 ) ) {
            int start;
            int end;

            GetVectorForKey(ent, "origin", org);

            start = dmodels[ i ].firstface;
            end = start + dmodels[ i ].numfaces;
            for( j = start; j < end; j++ ) {
                nolightface[ j ] = 300;
                faceoffset[ j ][ 0 ] = org[ 0 ];
                faceoffset[ j ][ 1 ] = org[ 1 ];
                faceoffset[ j ][ 2 ] = org[ 2 ];
            }
        }
    }
}

/*
 * =============
 *  LightWorld
 * =============
 */
void LightWorld (void)
{
    filebase = file_p = dlightdata;
    file_end = filebase + MAX_MAP_LIGHTING;

    RunThreadsOn (LightThread);

    lightdatasize = file_p - filebase;

    logprint ("lightdatasize: %i\n", lightdatasize);
}


/*
 * ==================
 * main
 * light modelfile
 * ==================
 */
int main (int argc, char **argv)
{
    int    i;
    double start;
    double end;
    char   source[1024];

    init_log("tomlite.log");
    logprint ("----- TomLite beta -----\n");

    for (i=1 ; i<argc ; i++) 
	{
        if (!strcmp(argv[i],"-threads")) 
		{
            numthreads = atoi (argv[i+1]);
            i++;
        } 
		else if (!strcmp(argv[i],"-extra")) 
		{
            extrasamples = true;
            logprint ("extra sampling enabled\n");
        } 
		else if (!strcmp(argv[i],"-dist")) 
		{
            scaledist = atof (argv[i+1]);
            i++;
        } 
		else if (!strcmp(argv[i],"-range")) 
		{
            rangescale = atof (argv[i+1]);
            i++;
        } 
		else if (!strcmp(argv[i],"-light")) 
		{
            worldminlight = atof (argv[i+1]);
            i++;
        } 
		else if (!strcmp(argv[i],"-compress")) 
		{
            compress_ents = true;
            logprint ("light entity compression enabled\n");
        } 
		else if (!strcmp(argv[i],"-nocount")) 
		{
            facecounter = true;
        } 
		else if (!strcmp(argv[i],"-colored")) 
		{
            colored = true;
			logprint (".lit output enabled\n");
        } 
		// Tomaz - Tint Begin
		else if (!strcmp(argv[i],"-tint")) 
		{
            tint = true;
        } 
		// Tomaz - Tint End
		else if (!strcmp(argv[i],"-nominlimit")) 
		{
            nominlimit = true;
        } 
		else if (argv[i][0] == '-')
            Error ("Unknown option \"%s\"", argv[i]);
        else
            break;
    }

    if (i != argc - 1)
        Error ("usage: tomlite [-colored] [-light num] [-extra] [-tint]\n"
               "             [-nocount] [-compress] [-nominlimit] bspfile");

    InitThreads ();

    start = I_FloatTime ();

    strcpy (source, argv[i]);
    StripExtension (source);
    DefaultExtension (source, ".bsp");

    LoadBSPFile (source);
    LoadEntities ();

    MakeTnodes (&dmodels[0]);

    FindFaceOffsets();
    LightWorld ();

    WriteEntitiesToString ();

	WriteBSPFile (source, BSPVERSION);

    end = I_FloatTime ();
    logprint ("%5.1f seconds elapsed\n", end-start);

    close_log();

    return 0;
}
